@@ -76,12 +76,26 @@ def qdump__std____1__complex(d, value):
 
 def qdump__std__deque(d, value):
     if d.isQnxTarget():
-        qdump__std__deque__QNX(d, value)
-        return
-    if d.isMsvcTarget():
-        qdump__std__deque__MSVC(d, value)
-        return
+        qdumpHelper__std__deque__qnx(d, value)
+    elif d.isMsvcTarget():
+        qdumpHelper__std__deque__msvc(d, value)
+    elif value.hasMember("_M_impl"):
+        qdumpHelper__std__deque__libstdcxx(d, value)
+    elif value.hasMember("__start_"):
+        qdumpHelper__std__deque__libcxx(d, value)
+    elif value.type.size() == 10 * d.ptrSize():
+        qdumpHelper__std__deque__libstdcxx(d, value)
+    elif value.type.size() == 6 * d.ptrSize():
+        qdumpHelper__std__deque__libcxx(d, value)
+    else:
+        qdumpHelper__std__deque__libstdcxx(d, value)
 
+
+def qdump__std____1__deque(d, value):
+    qdumpHelper__std__deque__libcxx(d, value)
+
+
+def qdumpHelper__std__deque__libstdcxx(d, value):
     innerType = value.type[0]
     innerSize = innerType.size()
     bufsize = 1
@@ -113,7 +127,7 @@ def qdump__std__deque(d, value):
                     pnode = newnode
 
 
-def qdump__std____1__deque(d, value):
+def qdumpHelper__std__deque__libcxx(d, value):
     mptr, mfirst, mbegin, mend, start, size = value.split("pppptt")
     d.check(0 <= size and size <= 1000 * 1000 * 1000)
     d.putItemCount(size)
@@ -129,7 +143,7 @@ def qdump__std____1__deque(d, value):
                 d.putSubItem(i, d.createValue(base + j * innerSize, innerType))
 
 
-def qdump__std__deque__QNX(d, value):
+def qdumpHelper__std__deque__qnx(d, value):
     innerType = value.type[0]
     innerSize = innerType.size()
     if innerSize <= 1:
@@ -166,7 +180,7 @@ def qdump__std__deque__QNX(d, value):
                 myoff += 1
 
 
-def qdump__std__deque__MSVC(d, value):
+def qdumpHelper__std__deque__msvc(d, value):
     innerType = value.type[0]
     innerSize = innerType.size()
     if innerSize <= 1:
@@ -228,7 +242,11 @@ def qdump__std__list(d, value):
 
 
 def qdump__std__list__QNX(d, value):
-    (proxy, head, size) = value.split("ppp")
+    try:
+        _ = value["_Mypair"]["_Myval2"]["_Myproxy"]
+        (proxy, head, size) = value.split("ppp")
+    except Exception:
+        (head, size) = value.split("pp")
     d.putItemCount(size, 1000)
 
     if d.isExpanded():
@@ -316,7 +334,11 @@ def qdump__std__map(d, value):
 
 
 def qdump_std__map__helper(d, value):
-    (proxy, head, size) = value.split("ppp")
+    try:
+        _ = value["_Mypair"]["_Myval2"]["_Myval2"]["_Myproxy"]
+        (proxy, head, size) = value.split("ppp")
+    except Exception:
+        (head, size) = value.split("pp")
     d.check(0 <= size and size <= 100 * 1000 * 1000)
     d.putItemCount(size)
     if d.isExpanded():
@@ -491,7 +513,11 @@ def qdump__std__set(d, value):
 
 
 def qdump__std__set__QNX(d, value):
-    (proxy, head, size) = value.split("ppp")
+    try:
+        _ = value["_Mypair"]["_Myval2"]["_Myval2"]["_Myproxy"]
+        (proxy, head, size) = value.split("ppp")
+    except Exception:
+        (head, size) = value.split("pp")
     d.check(0 <= size and size <= 100 * 1000 * 1000)
     d.putItemCount(size)
     if d.isExpanded():
@@ -617,7 +643,7 @@ def qdump__std____1__map(d, value):
 
         with Children(d, size, maxNumChild=1000):
             for (i, pair) in zip(d.childRange(), in_order_traversal(head)):
-                d.putPairItem(i, pair, 'key', 'value')
+                d.putPairItem(i, pair)
 
 
 def qform__std____1__multimap():
@@ -730,12 +756,12 @@ def qdumpHelper__std__string__QNX(d, value, charType, format):
 
 def qdumpHelper__std__string__MSVC(d, value, charType, format):
     try:
+        _ = value["_Mypair"]["_Myval2"]["_Myproxy"]
         (proxy, buffer, size, alloc) = value.split("p16spp")
-        d.check(0 <= size and size <= alloc and alloc <= 100 * 1000 * 1000)
-    except RuntimeError:
+    except Exception:
         proxy = None
         (buffer, size, alloc) = value.split("16spp")
-        d.check(0 <= size and size <= alloc and alloc <= 100 * 1000 * 1000)
+    d.check(0 <= size and size <= alloc and alloc <= 100 * 1000 * 1000)
     _BUF_SIZE = int(16 / charType.size())
     if _BUF_SIZE <= alloc:
         if proxy is None:
@@ -750,28 +776,54 @@ def qdumpHelper__std__string__MSVC(d, value, charType, format):
     d.putCharArrayHelper(data, size, charType, format)
 
 
+def qdump__std____1__string__alternateLayoutHelper(d, value):
+    try:
+        _d = value['__s']['__data_'].address()
+        alternateLayout = (_d - value.address()) == 0
+        if alternateLayout:
+            lastByte = value.split('b')[-1]
+            if int(lastByte & 0x80) == 0:
+                # Short/internal
+                size = value['__s']['__size_'].integer()
+                data = value.address()
+            else:
+                # Long/external
+                (data, size, _) = value.split('ppp')
+            return size, data
+        else:
+            return None, None
+    except:
+        return None, None
+
+
 def qdump__std____1__string(d, value):
-    firstByte = value.split('b')[0]
-    if int(firstByte & 1) == 0:
-        # Short/internal.
-        size = int(firstByte / 2)
-        data = value.address() + 1
-    else:
-        # Long/external.
-        (dummy, size, data) = value.split('ppp')
+    (size, data) = qdump__std____1__string__alternateLayoutHelper(d, value)
+    if size is None or data is None:
+        firstByte = value.split('b')[0]
+        if int(firstByte & 1) == 0:
+            # Short/internal.
+            size = int(firstByte / 2)
+            data = value.address() + 1
+        else:
+            # Long/external.
+            (_, size, data) = value.split('ppp')
+
     d.putCharArrayHelper(data, size, d.charType(), d.currentItemFormat())
     d.putType("std::string")
 
 
 def qdump__std____1__wstring(d, value):
-    firstByte = value.split('b')[0]
-    if int(firstByte & 1) == 0:
-        # Short/internal.
-        size = int(firstByte / 2)
-        data = value.address() + 4
-    else:
-        # Long/external.
-        (dummy, size, data) = value.split('ppp')
+    (size, data) = qdump__std____1__string__alternateLayoutHelper(d, value)
+    if size is None or data is None:
+        firstByte = value.split('b')[0]
+        if int(firstByte & 1) == 0:
+            # Short/internal.
+            size = int(firstByte / 2)
+            data = value.address() + 4
+        else:
+            # Long/external.
+            (_, size, data) = value.split('ppp')
+
     d.putCharArrayHelper(data, size, d.createType('wchar_t'))
     d.putType("std::wstring")
 
@@ -832,10 +884,9 @@ def qdump__std__pair(d, value):
     with Children(d):
         key = d.putSubItem('first', first)
         value = d.putSubItem('second', second)
-    d.putField('key', key.value)
-    if key.encoding is not None:
-        d.putField('keyencoded', key.encoding)
-    d.putValue(value.value, value.encoding)
+    key = key.value if key.encoding is None else "..."
+    value = value.value if value.encoding is None else "..."
+    d.putValue('(%s, %s)' % (key, value))
 
 
 def qform__std__unordered_map():
@@ -847,43 +898,58 @@ def qform__std____debug__unordered_map():
 
 
 def qdump__std__unordered_map(d, value):
-    if d.isQnxTarget() or d.isMsvcTarget():
+    if d.isQnxTarget():
         qdump__std__list__QNX(d, value["_List"])
         return
 
-    try:
-        # gcc ~= 4.7
-        size = value["_M_element_count"].integer()
-        start = value["_M_before_begin"]["_M_nxt"]
-    except:
+    if d.isMsvcTarget():
+        _list = value["_List"]
         try:
-            # libc++ (Mac)
-            size = value["_M_h"]["_M_element_count"].integer()
-            start = value["_M_h"]["_M_bbegin"]["_M_node"]["_M_nxt"]
+            _ = _list["_Mypair"]["_Myval2"]["_Myproxy"]
+            (_, start, size) = _list.split("ppp")
+        except Exception:
+            (start, size) = _list.split("pp")
+    else:
+        try:
+            # gcc ~= 4.7
+            size = value["_M_element_count"].integer()
+            start = value["_M_before_begin"]["_M_nxt"]
         except:
             try:
-                # gcc 4.9.1
+                # libc++ (Mac)
                 size = value["_M_h"]["_M_element_count"].integer()
-                start = value["_M_h"]["_M_before_begin"]["_M_nxt"]
+                start = value["_M_h"]["_M_bbegin"]["_M_node"]["_M_nxt"]
             except:
-                # gcc 4.6.2
-                size = value["_M_element_count"].integer()
-                start = value["_M_buckets"].dereference()
-                # FIXME: Pointer-aligned?
-                d.putItemCount(size)
-                # We don't know where the data is
-                d.putNumChild(0)
-                return
+                try:
+                    # gcc 4.9.1
+                    size = value["_M_h"]["_M_element_count"].integer()
+                    start = value["_M_h"]["_M_before_begin"]["_M_nxt"]
+                except:
+                    # gcc 4.6.2
+                    size = value["_M_element_count"].integer()
+                    start = value["_M_buckets"].dereference()
+                    # FIXME: Pointer-aligned?
+                    d.putItemCount(size)
+                    # We don't know where the data is
+                    d.putNumChild(0)
+                    return
 
     d.putItemCount(size)
     if d.isExpanded():
         keyType = value.type[0]
         valueType = value.type[1]
-        typeCode = 'p@{%s}@{%s}' % (value.type[0].name, value.type[1].name)
-        p = start.pointer()
+        if d.isMsvcTarget():
+            typeCode = 'pp@{%s}@{%s}' % (keyType.name, valueType.name)
+            p = d.extractPointer(start)
+        else:
+            typeCode = 'p@{%s}@{%s}' % (keyType.name, valueType.name)
+            p = start.pointer()
         with Children(d, size):
             for i in d.childRange():
-                p, pad, key, pad, val = d.split(typeCode, p)
+                if d.isMsvcTarget():
+                    p, _, _, key, _, val = d.split(typeCode, p)
+                else:
+                    p, _, key, _, val = d.split(typeCode, p)
                 d.putPairItem(i, (key, val))
 
 
@@ -1028,6 +1094,16 @@ def qdump__std____1__valarray(d, value):
     d.putItemCount(size)
     d.putPlotData(begin, size, innerType)
 
+def qdump__std__variant(d, value):
+    which = int(value["_M_index"])
+    type = d.templateArgument(value.type, which)
+    d.putValue("<%s:%s>" % (which, type.name))
+
+    d.putNumChild(1)
+    if d.isExpanded():
+        storage = value["_M_u"]["_M_first"]["_M_storage"]
+        with Children(d, 1):
+            d.putSubItem("value", storage.cast(type))
 
 def qform__std__vector():
     return [DisplayFormat.ArrayPlot]
@@ -1048,95 +1124,88 @@ def qedit__std__vector(d, value, data):
 
 def qdump__std__vector(d, value):
     if d.isQnxTarget() or d.isMsvcTarget():
-        qdumpHelper__std__vector__QNX(d, value)
+        qdumpHelper__std__vector__msvc(d, value)
+    elif value.hasMember("_M_impl"):
+        qdumpHelper__std__vector__libstdcxx(d, value)
+    elif value.hasMember("__begin_"):
+        qdumpHelper__std__vector__libcxx(d, value)
     else:
-        qdumpHelper__std__vector(d, value, False)
+        qdumpHelper__std__vector__libstdcxx(d, value)
 
 
-def qdumpHelper__std__vector(d, value, isLibCpp):
-    innerType = value.type[0]
-    isBool = innerType.name == 'bool'
-
-    if isBool:
-        if isLibCpp:
-            start = value["__begin_"].pointer()
-            size = value["__size_"].integer()
-            alloc = size
-        else:
-            start = value["_M_start"]["_M_p"].pointer()
-            soffset = value["_M_start"]["_M_offset"].integer()
-            finish = value["_M_finish"]["_M_p"].pointer()
-            foffset = value["_M_finish"]["_M_offset"].integer()
-            alloc = value["_M_end_of_storage"].pointer()
-            size = (finish - start) * 8 + foffset - soffset  # 8 is CHAR_BIT.
-    else:
-        if isLibCpp:
-            start = value["__begin_"].pointer()
-            finish = value["__end_"].pointer()
-            alloc = value["__end_cap_"].pointer()
-        else:
-            start = value["_M_start"].pointer()
-            finish = value["_M_finish"].pointer()
-            alloc = value["_M_end_of_storage"].pointer()
-        size = int((finish - start) / innerType.size())
-        d.check(finish <= alloc)
-        if size > 0:
-            d.checkPointer(start)
-            d.checkPointer(finish)
-            d.checkPointer(alloc)
-
-    d.check(0 <= size and size <= 1000 * 1000 * 1000)
-
-    d.putItemCount(size)
-    if isBool:
-        if d.isExpanded():
-            with Children(d, size, maxNumChild=10000, childType=innerType):
-                for i in d.childRange():
-                    q = start + int(i / 8)
-                    with SubItem(d, i):
-                        d.putValue((int(d.extractPointer(q)) >> (i % 8)) & 1)
-                        d.putType("bool")
-    else:
-        d.putPlotData(start, size, innerType)
-
-
-def qdumpHelper__std__vector__QNX(d, value):
-    innerType = value.type[0]
-    isBool = innerType.name == 'bool'
-    if isBool:
-        (proxy1, proxy2, start, last, end, size) = value.split("pppppi")
-    else:
-        (proxy, start, last, end) = value.split("pppp")
-        size = (last - start) // innerType.size()
-
-    try:
-        d.check(0 <= size and size <= 1000 * 1000 * 1000)
-        d.check(last <= end)
-    except RuntimeError:
-        if isBool:
-            (start, last, end, size) = value.split("pppi")
-        else:
-            (start, last, end) = value.split("ppp")
-            size = (last - start) // innerType.size()
-        d.check(0 <= size and size <= 1000 * 1000 * 1000)
-        d.check(last <= end)
-
+def qdumpHelper__std__vector__nonbool(d, start, finish, alloc, inner_type):
+    size = int((finish - start) / inner_type.size())
+    d.check(finish <= alloc)
     if size > 0:
         d.checkPointer(start)
-        d.checkPointer(last)
-        d.checkPointer(end)
+        d.checkPointer(finish)
+        d.checkPointer(alloc)
+    d.check(0 <= size and size <= 1000 * 1000 * 1000)
+    d.putItemCount(size)
+    d.putPlotData(start, size, inner_type)
 
+
+def qdumpHelper__std__vector__bool(d, start, size, inner_type):
+    d.check(0 <= size and size <= 1000 * 1000 * 1000)
     d.putItemCount(size)
     if d.isExpanded():
-        if isBool:
-            with Children(d, size, maxNumChild=10000, childType=innerType):
-                for i in d.childRange():
-                    q = start + int(i / 8)
-                    with SubItem(d, i):
-                        d.putValue((d.extractPointer(q) >> (i % 8)) & 1)
-                        d.putType("bool")
-        else:
-            d.putPlotData(start, size, innerType)
+        with Children(d, size, maxNumChild=10000, childType=inner_type):
+            for i in d.childRange():
+                q = start + int(i / 8)
+                with SubItem(d, i):
+                    d.putValue((int(d.extractPointer(q)) >> (i % 8)) & 1)
+                    d.putType("bool")
+
+
+def qdumpHelper__std__vector__libstdcxx(d, value):
+    inner_type = value.type[0]
+    if inner_type.name == "bool":
+        start = value["_M_start"]["_M_p"].pointer()
+        soffset = value["_M_start"]["_M_offset"].integer()
+        finish = value["_M_finish"]["_M_p"].pointer()
+        foffset = value["_M_finish"]["_M_offset"].integer()
+        alloc = value["_M_end_of_storage"].pointer()
+        size = (finish - start) * 8 + foffset - soffset  # 8 is CHAR_BIT.
+        qdumpHelper__std__vector__bool(d, start, size, inner_type)
+    else:
+        start = value["_M_start"].pointer()
+        finish = value["_M_finish"].pointer()
+        alloc = value["_M_end_of_storage"].pointer()
+        qdumpHelper__std__vector__nonbool(d, start, finish, alloc, inner_type)
+
+
+def qdumpHelper__std__vector__libcxx(d, value):
+    inner_type = value.type[0]
+    if inner_type.name == "bool":
+        start = value["__begin_"].pointer()
+        size = value["__size_"].integer()
+        qdumpHelper__std__vector__bool(d, start, size, inner_type)
+    else:
+        start = value["__begin_"].pointer()
+        finish = value["__end_"].pointer()
+        alloc = value["__end_cap_"].pointer()
+        qdumpHelper__std__vector__nonbool(d, start, finish, alloc, inner_type)
+
+
+def qdumpHelper__std__vector__msvc(d, value):
+    inner_type = value.type[0]
+    if inner_type.name == "bool":
+        try:
+            _ = value["_Myproxy"]
+            proxy1, proxy2, start, finish, alloc, size = value.split("pppppi")
+        except RuntimeError:
+            start, finish, alloc, size = value.split("pppi")
+        d.check(0 <= size and size <= 1000 * 1000 * 1000)
+        qdumpHelper__std__vector__bool(d, start, size, inner_type)
+    else:
+        try:
+            _ = value["_Mypair"]["_Myval2"]["_Myproxy"]
+            proxy, start, finish, alloc = value.split("pppp")
+        except RuntimeError:
+            start, finish, alloc = value.split("ppp")
+        size = (finish - start) // inner_type.size()
+        d.check(0 <= size and size <= 1000 * 1000 * 1000)
+        qdumpHelper__std__vector__nonbool(d, start, finish, alloc, inner_type)
 
 
 def qform__std____1__vector():
@@ -1144,7 +1213,7 @@ def qform__std____1__vector():
 
 
 def qdump__std____1__vector(d, value):
-    qdumpHelper__std__vector(d, value, True)
+    qdumpHelper__std__vector__libcxx(d, value)
 
 
 def qform__std____debug__vector():
@@ -1250,9 +1319,9 @@ def qform__std____cxx11__wstring():
 
 def qdump__std____1__basic_string(d, value):
     innerType = value.type[0].name
-    if innerType == "char":
+    if innerType in ("char", "char8_t", "char16_t"):
         qdump__std____1__string(d, value)
-    elif innerType == "wchar_t":
+    elif innerType in ("wchar_t", "char32_t"):
         qdump__std____1__wstring(d, value)
     else:
         d.warn("UNKNOWN INNER TYPE %s" % innerType)
